@@ -1,6 +1,6 @@
 /*
  Copyright 2018 Google Inc.
- Copyright (c) 2023-2025 LunarG, Inc.
+ Copyright (c) 2023-2026 LunarG, Inc.
 
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
@@ -82,6 +82,7 @@ static const std::unordered_map<std::string, DumpShaders> kDumpShadersValues{
     {"all", DumpShaders::kAll},
 };
 
+const char* kTriggerWatchdogTimeout = "trigger_watchdog_timeout";
 const char* kWatchdogTimeout = "watchdog_timeout_ms";
 const char* kDumpAllCommandBuffers = "dump_all_command_buffers";
 const char* kTrackSemaphores = "track_semaphores";
@@ -166,6 +167,7 @@ void Settings::Print(YAML::Emitter& os) const {
     os << YAML::Key << settings::kDumpCommandBuffers << YAML::Value << dump_command_buffers;
     os << YAML::Key << settings::kDumpCommands << YAML::Value << dump_commands;
     os << YAML::Key << settings::kDumpShaders << YAML::Value << dump_shaders;
+    os << YAML::Key << settings::kTriggerWatchdogTimeout << YAML::Value << trigger_watchdog_timer;
     os << YAML::Key << settings::kWatchdogTimeout << YAML::Value << watchdog_timer_ms;
     os << YAML::Key << settings::kTrackSemaphores << YAML::Value << track_semaphores;
     os << YAML::Key << settings::kTraceAllSemaphores << YAML::Value << trace_all_semaphores;
@@ -210,16 +212,18 @@ Context::Context(const VkInstanceCreateInfo* pCreateInfo, const VkAllocationCall
     settings_.emplace(layer_setting_set, logger_);
 
     // output path
+    std::filesystem::path default_output_path = System::GetOutputBasePath();
+    default_output_path /= "cdl";
     {
         if (!settings_->output_path.empty()) {
             output_path_ = settings_->output_path;
         } else {
-            output_path_ = System::GetOutputBasePath();
-            output_path_ /= "cdl";
+            output_path_ = default_output_path;
         }
 
         base_output_path_ = output_path_;
         output_path_ /= start_time_str;
+        default_output_path /= start_time_str;
     }
     // logging
     {
@@ -271,6 +275,20 @@ Context::Context(const VkInstanceCreateInfo* pCreateInfo, const VkAllocationCall
                         path = output_path_ / log_file;
                     }
                 }
+
+                if (path.has_parent_path()) {
+                    try {
+                        // On Windows, if the path include an invalid drive, then filesystem crash...
+                        // We need to use an exception to capture this problem
+                        std::filesystem::create_directories(path.parent_path());
+                    } catch (std::exception& e) {
+                        std::cerr << "Failed to open log file: " << path << " - " << e.what() << std::endl;
+                        // We fallback using the default path
+                        output_path_ = default_output_path;
+                        path = default_output_path / log_file;
+                    }
+                }
+
                 logger_.OpenLogFile(path);
             }
         }
@@ -425,6 +443,16 @@ static DeviceExtensionsPresent DecodeExtensionStrings(uint32_t count, const char
     }
     return extensions;
 }
+template <typename T>
+void Context::QueryFeature(VkPhysicalDevice physicalDevice, T *feature) {
+    auto features2 = vku::InitStruct<VkPhysicalDeviceFeatures2>(feature);
+    if (modified_create_info_.pApplicationInfo &&
+        modified_create_info_.pApplicationInfo->apiVersion >= VK_API_VERSION_1_1) {
+        Dispatch().GetPhysicalDeviceFeatures2(physicalDevice, &features2);
+    } else {
+        Dispatch().GetPhysicalDeviceFeatures2KHR(physicalDevice, &features2);
+    }
+}
 
 const VkDeviceCreateInfo* Context::GetModifiedDeviceCreateInfo(VkPhysicalDevice physicalDevice,
                                                                const VkDeviceCreateInfo* pCreateInfo) {
@@ -448,6 +476,10 @@ const VkDeviceCreateInfo* Context::GetModifiedDeviceCreateInfo(VkPhysicalDevice
     device_ci->original.initialize(pCreateInfo);
     device_ci->modified = device_ci->original;
 
+    const uint32_t api_version = modified_create_info_.pApplicationInfo
+                                     ? modified_create_info_.pApplicationInfo->apiVersion
+                                     : VK_API_VERSION_1_0;
+
     // If an important extension is not enabled by default, try to enable it if it is present
     if (extensions_present.nv_device_diagnostic_checkpoints) {
         if (!extensions_enabled.nv_device_diagnostic_checkpoints) {
@@ -463,11 +495,18 @@ const VkDeviceCreateInfo* Context::GetModifiedDeviceCreateInfo(VkPhysicalDevice
             vku::AddExtension(device_ci->modified, VK_AMD_BUFFER_MARKER_EXTENSION_NAME);
         }
         if (extensions_present.amd_coherent_memory) {
-            if (!extensions_enabled.amd_coherent_memory) {
-                extensions_enabled.amd_coherent_memory = true;
-                auto amd_device_coherent = vku::InitStruct<VkPhysicalDeviceCoherentMemoryFeaturesAMD>(nullptr, VK_TRUE);
-                vku::AddToPnext(device_ci->modified, amd_device_coherent);
-                vku::AddExtension(device_ci->modified, VK_AMD_DEVICE_COHERENT_MEMORY_EXTENSION_NAME);
+            auto amd_device_coherent = vku::InitStruct<VkPhysicalDeviceCoherentMemoryFeaturesAMD>(nullptr);
+            QueryFeature(physicalDevice, &amd_device_coherent);
+            if (amd_device_coherent.deviceCoherentMemory) {
+                auto *existing = vku::FindStructInPNextChain<VkPhysicalDeviceCoherentMemoryFeaturesAMD>(&device_ci->modified);
+                if (existing) {
+                    existing->deviceCoherentMemory = amd_device_coherent.deviceCoherentMemory;
+                } else {
+                    vku::AddToPnext(device_ci->modified, amd_device_coherent);
+                    if (!extensions_enabled.amd_coherent_memory) {
+                        vku::AddExtension(device_ci->modified, VK_AMD_DEVICE_COHERENT_MEMORY_EXTENSION_NAME);
+                    }
+                }
             }
         } else {
             Log().Warning("No VK_AMD_device_coherent_memory extension, results may not be as accurate as possible.");
@@ -482,52 +521,64 @@ const VkDeviceCreateInfo* Context::GetModifiedDeviceCreateInfo(VkPhysicalDevice
     }
     if (extensions_present.ext_device_fault) {
         if (!extensions_enabled.ext_device_fault) {
-            extensions_enabled.ext_device_fault = true;
-            // Query the feature so that we know if vendor data is supported
             auto ext_device_fault = vku::InitStruct<VkPhysicalDeviceFaultFeaturesEXT>(nullptr);
-            auto features2 = vku::InitStruct<VkPhysicalDeviceFeatures2>(&ext_device_fault);
-            if (modified_create_info_.pApplicationInfo &&
-                modified_create_info_.pApplicationInfo->apiVersion >= VK_API_VERSION_1_1) {
-                Dispatch().GetPhysicalDeviceFeatures2(physicalDevice, &features2);
-            } else {
-                Dispatch().GetPhysicalDeviceFeatures2KHR(physicalDevice, &features2);
+            QueryFeature(physicalDevice, &ext_device_fault);
+            if (ext_device_fault.deviceFault) {
+                auto *existing = vku::FindStructInPNextChain<VkPhysicalDeviceFaultFeaturesEXT>(&device_ci->modified);
+                if (existing) {
+                    existing->deviceFault = ext_device_fault.deviceFault;
+                    existing->deviceFaultVendorBinary = ext_device_fault.deviceFaultVendorBinary;
+                } else {
+                    vku::AddToPnext(device_ci->modified, ext_device_fault);
+                    if (!extensions_enabled.ext_device_fault) {
+                        vku::AddExtension(device_ci->modified, VK_EXT_DEVICE_FAULT_EXTENSION_NAME);
+                    }
+                }
             }
-
-            vku::AddToPnext(device_ci->modified, ext_device_fault);
-            vku::AddExtension(device_ci->modified, VK_EXT_DEVICE_FAULT_EXTENSION_NAME);
         }
     } else {
         Log().Warning("No VK_EXT_device_fault extension, vendor-specific crash dumps will not be available.");
     }
     if (extensions_present.ext_device_address_binding_report) {
-        if (!extensions_enabled.ext_device_address_binding_report) {
-            extensions_enabled.ext_device_address_binding_report = true;
-            auto ext_device_address_binding_report =
-                vku::InitStruct<VkPhysicalDeviceAddressBindingReportFeaturesEXT>(nullptr, VK_TRUE);
-            vku::AddToPnext(device_ci->modified, ext_device_address_binding_report);
-            vku::AddExtension(device_ci->modified, VK_EXT_DEVICE_ADDRESS_BINDING_REPORT_EXTENSION_NAME);
+        auto ext_dbar = vku::InitStruct<VkPhysicalDeviceAddressBindingReportFeaturesEXT>(nullptr);
+        QueryFeature(physicalDevice, &ext_dbar);
+        if (ext_dbar.reportAddressBinding) {
+            auto *existing = vku::FindStructInPNextChain<VkPhysicalDeviceAddressBindingReportFeaturesEXT>(&device_ci->modified);
+            if (existing) {
+                existing->reportAddressBinding = ext_dbar.reportAddressBinding;
+            } else {
+                vku::AddToPnext(device_ci->modified, ext_dbar);
+                if (!extensions_enabled.ext_device_address_binding_report) {
+                    vku::AddExtension(device_ci->modified, VK_EXT_DEVICE_ADDRESS_BINDING_REPORT_EXTENSION_NAME);
+                }
+            }
         }
     } else {
         Log().Warning(
             "No VK_EXT_device_address_binding_report extension, DeviceAddress information will not be available.");
     }
     if (extensions_present.khr_timeline_semaphore) {
-        if (!extensions_enabled.khr_timeline_semaphore) {
-            extensions_enabled.khr_timeline_semaphore = true;
-            auto* vulkan12_features =
-                vku::FindStructInPNextChain<VkPhysicalDeviceVulkan12Features>(&device_ci->modified);
+        auto khr_timeline_semaphore = vku::InitStruct<VkPhysicalDeviceTimelineSemaphoreFeaturesKHR>(nullptr);
+        QueryFeature(physicalDevice, &khr_timeline_semaphore);
+
+        if (khr_timeline_semaphore.timelineSemaphore) {
+            auto* vulkan12_features = vku::FindStructInPNextChain<VkPhysicalDeviceVulkan12Features>(&device_ci->modified);
+            auto* existing = vku::FindStructInPNextChain<VkPhysicalDeviceTimelineSemaphoreFeaturesKHR>(&device_ci->modified);
             if (vulkan12_features) {
                 vulkan12_features->timelineSemaphore = VK_TRUE;
+            } else if (existing) {
+                existing->timelineSemaphore = VK_TRUE;
             } else {
-                auto khr_timeline_semaphore =
-                    vku::InitStruct<VkPhysicalDeviceTimelineSemaphoreFeaturesKHR>(nullptr, VK_TRUE);
                 vku::AddToPnext(device_ci->modified, khr_timeline_semaphore);
-                vku::AddExtension(device_ci->modified, VK_KHR_TIMELINE_SEMAPHORE_EXTENSION_NAME);
+                if (api_version < VK_API_VERSION_1_2 && !extensions_enabled.khr_timeline_semaphore) {
+                    vku::AddExtension(device_ci->modified, VK_KHR_TIMELINE_SEMAPHORE_EXTENSION_NAME);
+                }
             }
         }
     } else {
-        Log().Warning(
-            "No VK_EXT_device_address_binding_report extension, DeviceAddress information will not be available.");
+        Log().Error(
+            "No VK_KHR_timeline_semaphore extension, Vulkan 1.2 or VK_KHR_timeline_semaphore are quired to track queue "
+            "progress, enabling early device lost detection.");
     }
 
     // save the raw ptr before std::move of the std::unique_ptr
@@ -691,40 +742,6 @@ std::ofstream Context::OpenDumpFile() {
 // Define pre / post intercepted commands
 // =============================================================================
 
-static VKAPI_ATTR VkBool32 VKAPI_CALL MessengerCallback(VkDebugUtilsMessageSeverityFlagBitsEXT severity,
-                                                        VkDebugUtilsMessageTypeFlagsEXT types,
-                                                        const VkDebugUtilsMessengerCallbackDataEXT* pCallbackData,
-                                                        void* pUserData) {
-    if (nullptr == pCallbackData || nullptr == pUserData) {
-        return VK_FALSE;
-    }
-    Context* context = reinterpret_cast<Context*>(pUserData);
-    const auto* mem_info = vku::FindStructInPNextChain<VkDeviceAddressBindingCallbackDataEXT>(pCallbackData->pNext);
-    if (!mem_info) {
-        return VK_FALSE;
-    }
-    context->MemoryBindEvent(*mem_info, *pCallbackData->pObjects);
-    return VK_FALSE;
-}
-
-void Context::MemoryBindEvent(const VkDeviceAddressBindingCallbackDataEXT& mem_info,
-                              const VkDebugUtilsObjectNameInfoEXT& object) {
-    DeviceAddressRecord rec{mem_info.baseAddress,
-                            mem_info.size,
-                            mem_info.flags,
-                            mem_info.bindingType,
-                            object.objectType,
-                            object.objectHandle,
-                            object.pObjectName ? object.pObjectName : "",
-                            std::chrono::high_resolution_clock::now()};
-    auto devs = GetAllDevices();
-
-    bool multi_device = devs.size() > 1;
-    for (auto& dev : devs) {
-        dev->MemoryBindEvent(rec, multi_device);
-    }
-}
-
 VkResult Context::PostCreateInstance(const VkInstanceCreateInfo* pCreateInfo, const VkAllocationCallbacks* pAllocator,
                                      VkInstance* pInstance, VkResult result) {
     vk_instance_ = *pInstance;
@@ -743,31 +760,9 @@ VkResult Context::PostCreateInstance(const VkInstanceCreateInfo* pCreateInfo, co
         application_info_->engineVersion = pCreateInfo->pApplicationInfo->engineVersion;
         application_info_->apiVersion = pCreateInfo->pApplicationInfo->apiVersion;
     }
-    // This messenger is for messages we recieve from the ICD for device address binding events.
-    if (VK_NULL_HANDLE == utils_messenger_ && instance_dispatch_table_.CreateDebugUtilsMessengerEXT) {
-        VkDebugUtilsMessengerCreateInfoEXT messenger_create_info = {
-            VK_STRUCTURE_TYPE_DEBUG_UTILS_MESSENGER_CREATE_INFO_EXT,
-            nullptr,
-            0,
-            VK_DEBUG_UTILS_MESSAGE_SEVERITY_ERROR_BIT_EXT | VK_DEBUG_UTILS_MESSAGE_SEVERITY_WARNING_BIT_EXT |
-                VK_DEBUG_UTILS_MESSAGE_SEVERITY_INFO_BIT_EXT,
-            VK_DEBUG_UTILS_MESSAGE_TYPE_DEVICE_ADDRESS_BINDING_BIT_EXT,
-            &MessengerCallback,
-            this,
-        };
-        instance_dispatch_table_.CreateDebugUtilsMessengerEXT(vk_instance_, &messenger_create_info, nullptr,
-                                                              &utils_messenger_);
-    }
     return result;
 }
 
-void Context::PreDestroyInstance(VkInstance instance, const VkAllocationCallbacks* pAllocator) {
-    if (VK_NULL_HANDLE != utils_messenger_) {
-        instance_dispatch_table_.DestroyDebugUtilsMessengerEXT(vk_instance_, utils_messenger_, nullptr);
-        utils_messenger_ = VK_NULL_HANDLE;
-    }
-}
-
 VkResult Context::PostCreateDevice(VkPhysicalDevice physicalDevice, const VkDeviceCreateInfo* pCreateInfo,
                                    const VkAllocationCallbacks* pAllocator, VkDevice* pDevice, VkResult callResult) {
     if (callResult != VK_SUCCESS) return callResult;
@@ -1284,8 +1279,11 @@ VkResult Context::PostWaitSemaphores(VkDevice device, const VkSemaphoreWaitInfoK
 
 VkResult Context::PostGetSemaphoreCounterValue(VkDevice device, VkSemaphore semaphore, uint64_t* pValue,
                                                VkResult result) {
+    auto device_state = GetDevice(device);
+    if (!device_state->UpdateIdleState()) {
+        result = VK_ERROR_DEVICE_LOST;
+    }
     if (IsVkError(result)) {
-        auto device_state = GetDevice(device);
         device_state->DeviceFault();
     }
     return result;
