@@ -1,6 +1,6 @@
 /*
  Copyright 2018 Google Inc.
- Copyright 2023-2024 LunarG, Inc.
+ Copyright 2023-2026 LunarG, Inc.
 
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
@@ -30,6 +30,32 @@
 
 namespace crash_diagnostic_layer {
 
+static VKAPI_ATTR VkBool32 VKAPI_CALL MessengerCallback(VkDebugUtilsMessageSeverityFlagBitsEXT severity,
+                                                        VkDebugUtilsMessageTypeFlagsEXT types,
+                                                        const VkDebugUtilsMessengerCallbackDataEXT* pCallbackData,
+                                                        void* pUserData) {
+    if (nullptr == pCallbackData || nullptr == pUserData) {
+        return VK_FALSE;
+    }
+    Device* dev = reinterpret_cast<Device*>(pUserData);
+    const auto* mem_info = vku::FindStructInPNextChain<VkDeviceAddressBindingCallbackDataEXT>(pCallbackData->pNext);
+    if (!mem_info) {
+        return VK_FALSE;
+    }
+    const auto& object = *pCallbackData->pObjects;
+    DeviceAddressRecord rec{mem_info->baseAddress,
+                            mem_info->size,
+                            mem_info->flags,
+                            mem_info->bindingType,
+                            object.objectType,
+                            object.objectHandle,
+                            object.pObjectName ? object.pObjectName : "",
+                            std::chrono::high_resolution_clock::now()};
+    dev->MemoryBindEvent(rec, false);
+
+    return VK_FALSE;
+}
+
 Device::Device(Context& context, VkPhysicalDevice vk_gpu, VkDevice device, DeviceExtensionsPresent& extensions_present,
                std::unique_ptr<DeviceCreateInfo> device_create_info)
     : context_(context),
@@ -59,12 +85,31 @@ Device::Device(Context& context, VkPhysicalDevice vk_gpu, VkDevice device, Devic
     if (context_.GetSettings().track_semaphores) {
         semaphore_tracker_ = std::make_unique<SemaphoreTracker>(*this);
     }
-    if (context_.GetSettings().watchdog_timer_ms > 0) {
+    if (context_.GetSettings().trigger_watchdog_timer) {
         watchdog_.Start();
     }
+    // This messenger is for messages we recieve from the ICD for device address binding events.
+    if (extensions_present_.ext_device_address_binding_report && context.Dispatch().CreateDebugUtilsMessengerEXT) {
+        VkDebugUtilsMessengerCreateInfoEXT messenger_create_info = {
+            VK_STRUCTURE_TYPE_DEBUG_UTILS_MESSENGER_CREATE_INFO_EXT,
+            nullptr,
+            0,
+            VK_DEBUG_UTILS_MESSAGE_SEVERITY_ERROR_BIT_EXT | VK_DEBUG_UTILS_MESSAGE_SEVERITY_WARNING_BIT_EXT |
+                VK_DEBUG_UTILS_MESSAGE_SEVERITY_INFO_BIT_EXT,
+            VK_DEBUG_UTILS_MESSAGE_TYPE_DEVICE_ADDRESS_BINDING_BIT_EXT,
+            &MessengerCallback,
+            this,
+        };
+        context.Dispatch().CreateDebugUtilsMessengerEXT(context.GetInstance(), &messenger_create_info, nullptr,
+                                                        &utils_messenger_);
+    }
 }
 
 void Device::Destroy() {
+    if (VK_NULL_HANDLE != utils_messenger_) {
+        context_.Dispatch().DestroyDebugUtilsMessengerEXT(context_.GetInstance(), utils_messenger_, nullptr);
+        utils_messenger_ = VK_NULL_HANDLE;
+    }
     for (auto& item : queues_) {
         item.second->Destroy();
     }
@@ -105,16 +150,17 @@ void Device::DumpCommandBuffers(YAML::Emitter& os) const {
         auto p_cmd = GetCommandBuffer(cb);
         if (p_cmd && p_cmd->IsPrimaryCommandBuffer()) {
             bool dump_this_cb = false;
+            auto cb_state = p_cmd->GetCommandBufferState();
             switch (dump_cbs) {
                 case DumpCommands::kAll:
                     dump_this_cb = true;
                     break;
                 case DumpCommands::kRunning:
-                    dump_this_cb = p_cmd->GetCommandBufferState() == CommandBufferState::kSubmittedExecutionIncomplete;
+                    dump_this_cb = cb_state == CommandBufferState::kIncomplete
+                                   || cb_state == CommandBufferState::kMaybeComplete;
                     break;
                 case DumpCommands::kPending:
-                    dump_this_cb = p_cmd->WasSubmittedToQueue() &&
-                                   p_cmd->GetCommandBufferState() != CommandBufferState::kSubmittedExecutionNotStarted;
+                    dump_this_cb = p_cmd->WasSubmittedToQueue();
                     break;
             }
             if (dump_this_cb) {
@@ -198,7 +244,10 @@ void Device::DumpCommandBufferStateOnScreen(CommandBuffer* p_cmd, YAML::Emitter&
 bool Device::ValidateCommandBufferNotInUse(CommandBuffer* p_cmd, YAML::Emitter& os) {
     assert(p_cmd);
     if (!HangDetected()) {
-        if (p_cmd->HasCheckpoints() && p_cmd->WasSubmittedToQueue() && !p_cmd->CompletedExecution()) {
+        auto cb_state = p_cmd->GetCommandBufferState();
+        if (cb_state == CommandBufferState::kSubmitted ||
+            cb_state == CommandBufferState::kNotStarted ||
+            cb_state == CommandBufferState::kIncomplete) {
             DumpCommandBufferStateOnScreen(p_cmd, os);
             return false;
         }
@@ -579,24 +628,26 @@ void Device::DumpDeviceFaultInfo(YAML::Emitter& os) const {
             os << YAML::Key << "type" << YAML::Value << address_fault_strings[info.addressType];
             os << YAML::Key << "begin" << YAML::Value << Uint64ToStr(range.begin);
             os << YAML::Key << "end" << YAML::Value << Uint64ToStr(range.end);
-
-            auto lower = address_map_.lower_bound(range);
-            auto upper = address_map_.upper_bound(range);
-            if (lower != upper) {
-                os << YAML::Key << "matchingAddressRecords" << YAML::Value << YAML::BeginSeq;
-                for (auto iter = lower; iter != upper; ++iter) {
-                    DumpAddressRecord(os, iter->first, iter->second);
+            {
+                std::lock_guard<std::mutex> lock(address_mutex_);
+                auto lower = address_map_.lower_bound(range);
+                auto upper = address_map_.upper_bound(range);
+                if (lower != upper) {
+                    os << YAML::Key << "matchingAddressRecords" << YAML::Value << YAML::BeginSeq;
+                    for (auto iter = lower; iter != upper; ++iter) {
+                        DumpAddressRecord(os, iter->first, iter->second);
+                    }
+                    os << YAML::EndSeq;
+                }
+                if (lower != address_map_.begin()) {
+                    --lower;
+                    os << YAML::Key << "priorAddressRecord" << YAML::Value;
+                    DumpAddressRecord(os, lower->first, lower->second);
+                }
+                if (upper != address_map_.end()) {
+                    os << YAML::Key << "nextAddressRecord" << YAML::Value;
+                    DumpAddressRecord(os, upper->first, upper->second);
                 }
-                os << YAML::EndSeq;
-            }
-            if (lower != address_map_.begin()) {
-                --lower;
-                os << YAML::Key << "priorAddressRecord" << YAML::Value;
-                DumpAddressRecord(os, lower->first, lower->second);
-            }
-            if (upper != address_map_.end()) {
-                os << YAML::Key << "nextAddressRecord" << YAML::Value;
-                DumpAddressRecord(os, upper->first, upper->second);
             }
             os << YAML::EndMap;
         }
@@ -633,6 +684,7 @@ void Device::MemoryBindEvent(const DeviceAddressRecord& rec, bool multi_device)
     // up the vulkan handle. This will be slow.
     assert(!multi_device);
 
+    std::lock_guard<std::mutex> lock(address_mutex_);
     vku::sparse::range<VkDeviceAddress> range(rec.base, rec.base + rec.size);
     address_map_.overwrite_range(std::make_pair(range, rec));
 }
