@@ -1,6 +1,6 @@
 /*
  Copyright 2020 Google Inc.
- Copyright 2023-2024 LunarG, Inc.
+ Copyright 2023-2026 LunarG, Inc.
 
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
@@ -196,16 +196,12 @@ bool Queue::UpdateIdleState() {
                               device_.GetObjectInfo((uint64_t)vk_queue_).c_str(), submit_info.start_seq,
                               submit_info.end_seq);
                 submit_info.state = kRunning;
-                bool found_running_cb = false;
-                for (auto pos = submit_info.command_buffers.rbegin(); pos != submit_info.command_buffers.rend();
-                     ++pos) {
-                    auto cmd = crash_diagnostic_layer::GetCommandBuffer(*pos);
-                    if (cmd) {
-                        if (found_running_cb) {
-                            cmd->SetCompleted();
-                        } else if (cmd->GetCommandBufferState() == CommandBufferState::kSubmittedExecutionIncomplete) {
-                            found_running_cb = true;
-                        }
+                for (auto& cb : submit_info.command_buffers) {
+                    auto cmd = crash_diagnostic_layer::GetCommandBuffer(cb);
+                    if (cmd && completed_seq >= cmd->GetQueueSeq()) {
+                        Log().Verbose("%s cb completed: %lld queue_seq: %lld",
+                                      device_.GetObjectInfo((uint64_t)cb).c_str(), completed_seq, cmd->GetQueueSeq());
+                        cmd->SetCompleted();
                     }
                 }
                 break;
@@ -393,6 +389,7 @@ void Queue::Print(YAML::Emitter& os) {
                     value << device_.GetObjectInfo((uint64_t)cb);
                     auto cmd = crash_diagnostic_layer::GetCommandBuffer(cb);
                     if (cmd) {
+                        cmd->UpdateStateFromCheckpoints();
                         value << " " << cmd->PrintCommandBufferState() << " " << cmd->GetQueueSeq();
                     }
                     os << value.str();
@@ -650,17 +647,6 @@ VkResult Queue::Submit(uint32_t submitCount, const VkSubmitInfo* pSubmits, VkFen
             }
             vku::FreePnextChain(p_next);
         }
-        if (NeedSignalSubmission(pSubmits[i])) {
-            auto submit = vku::InitStruct<VkSubmitInfo>();
-            submit.pNext = MakeSignalSubmissionPnext(pSubmits[i].pNext);
-            submit.signalSemaphoreCount = pSubmits[i].signalSemaphoreCount;
-            submit.pSignalSemaphores = pSubmits[i].pSignalSemaphores;
-            if (result == VK_SUCCESS) {
-                result = device_.Dispatch().QueueSubmit(vk_queue_, 1, &submit, VK_NULL_HANDLE);
-            }
-            vku::FreePnextChain(submit.pNext);
-        }
-
         submit_info.end_seq = ++submit_seq_;
         {
             auto timeline_values = vku::InitStruct<VkTimelineSemaphoreSubmitInfo>();
@@ -673,6 +659,17 @@ VkResult Queue::Submit(uint32_t submitCount, const VkSubmitInfo* pSubmits, VkFen
                 result = device_.Dispatch().QueueSubmit(vk_queue_, 1, &submit, VK_NULL_HANDLE);
             }
         }
+        if (NeedSignalSubmission(pSubmits[i])) {
+            auto submit = vku::InitStruct<VkSubmitInfo>();
+            submit.pNext = MakeSignalSubmissionPnext(pSubmits[i].pNext);
+            submit.signalSemaphoreCount = pSubmits[i].signalSemaphoreCount;
+            submit.pSignalSemaphores = pSubmits[i].pSignalSemaphores;
+            if (result == VK_SUCCESS) {
+                result = device_.Dispatch().QueueSubmit(vk_queue_, 1, &submit, VK_NULL_HANDLE);
+            }
+            vku::FreePnextChain(submit.pNext);
+        }
+
         if (trace_all_semaphores_) {
             LogSubmitInfoSemaphores(submit_info);
         }
@@ -746,15 +743,6 @@ VkResult Queue::Submit2(uint32_t submitCount, const VkSubmitInfo2* pSubmits, VkF
                 result = QueueSubmit2(vk_queue_, 1, &submit, VK_NULL_HANDLE);
             }
         }
-        if (pSubmits[i].signalSemaphoreInfoCount > 0) {
-            auto submit = vku::InitStruct<VkSubmitInfo2>();
-            submit.signalSemaphoreInfoCount = pSubmits[i].signalSemaphoreInfoCount;
-            submit.pSignalSemaphoreInfos = pSubmits[i].pSignalSemaphoreInfos;
-            if (result == VK_SUCCESS) {
-                result = QueueSubmit2(vk_queue_, 1, &submit, VK_NULL_HANDLE);
-            }
-        }
-
         submit_info.end_seq = ++submit_seq_;
         {
             auto signal_info = vku::InitStruct<VkSemaphoreSubmitInfo>();
@@ -770,6 +758,15 @@ VkResult Queue::Submit2(uint32_t submitCount, const VkSubmitInfo2* pSubmits, VkF
                 break;
             }
         }
+        if (pSubmits[i].signalSemaphoreInfoCount > 0) {
+            auto submit = vku::InitStruct<VkSubmitInfo2>();
+            submit.signalSemaphoreInfoCount = pSubmits[i].signalSemaphoreInfoCount;
+            submit.pSignalSemaphoreInfos = pSubmits[i].pSignalSemaphoreInfos;
+            if (result == VK_SUCCESS) {
+                result = QueueSubmit2(vk_queue_, 1, &submit, VK_NULL_HANDLE);
+            }
+        }
+
         if (trace_all_semaphores_) {
             LogSubmitInfoSemaphores(submit_info);
         }
@@ -819,17 +816,27 @@ VkResult Queue::BindSparse(uint32_t bindInfoCount, const VkBindSparseInfo* pBind
             vku::FreePnextChain(bind_info.pNext);
         }
         if (result == VK_SUCCESS) {
-            result = device_.Dispatch().QueueBindSparse(vk_queue_, 1, &pBindInfos[i], VK_NULL_HANDLE);
-        }
-        if (pBindInfos[i].signalSemaphoreCount > 0) {
-            auto bind_info = vku::InitStruct<VkBindSparseInfo>();
-            bind_info.pNext = MakeSignalSubmissionPnext(pBindInfos[i].pNext);
-            bind_info.signalSemaphoreCount = pBindInfos[i].signalSemaphoreCount;
-            bind_info.pSignalSemaphores = pBindInfos[i].pSignalSemaphores;
-            if (result == VK_SUCCESS) {
-                result = device_.Dispatch().QueueBindSparse(vk_queue_, 1, &bind_info, VK_NULL_HANDLE);
-            }
-            vku::FreePnextChain(bind_info.pNext);
+            uint64_t bind_seq = ++submit_seq_;
+            void* p_next = MakeMiddleSubmissionPnext(pBindInfos[i].pNext);
+
+            auto timeline_values = vku::InitStruct<VkTimelineSemaphoreSubmitInfo>();
+            timeline_values.signalSemaphoreValueCount = 1;
+            timeline_values.pSignalSemaphoreValues = &bind_seq;
+            timeline_values.pNext = p_next;
+
+            auto bind_info = vku::InitStruct<VkBindSparseInfo>(&timeline_values);
+            bind_info.signalSemaphoreCount = 1;
+            bind_info.pSignalSemaphores = &submit_sem_;
+
+            bind_info.bufferBindCount = pBindInfos[i].bufferBindCount;
+            bind_info.pBufferBinds = pBindInfos[i].pBufferBinds;
+            bind_info.imageOpaqueBindCount = pBindInfos[i].imageOpaqueBindCount;
+            bind_info.pImageOpaqueBinds = pBindInfos[i].pImageOpaqueBinds;
+            bind_info.imageBindCount = pBindInfos[i].imageBindCount;
+            bind_info.pImageBinds = pBindInfos[i].pImageBinds;
+
+            result = device_.Dispatch().QueueBindSparse(vk_queue_, 1, &bind_info, VK_NULL_HANDLE);
+            vku::FreePnextChain(p_next);
         }
         submit_info.end_seq = ++submit_seq_;
         {
@@ -843,6 +850,16 @@ VkResult Queue::BindSparse(uint32_t bindInfoCount, const VkBindSparseInfo* pBind
                 result = device_.Dispatch().QueueSubmit(vk_queue_, 1, &submit, VK_NULL_HANDLE);
             }
         }
+        if (pBindInfos[i].signalSemaphoreCount > 0) {
+            auto bind_info = vku::InitStruct<VkBindSparseInfo>();
+            bind_info.pNext = MakeSignalSubmissionPnext(pBindInfos[i].pNext);
+            bind_info.signalSemaphoreCount = pBindInfos[i].signalSemaphoreCount;
+            bind_info.pSignalSemaphores = pBindInfos[i].pSignalSemaphores;
+            if (result == VK_SUCCESS) {
+                result = device_.Dispatch().QueueBindSparse(vk_queue_, 1, &bind_info, VK_NULL_HANDLE);
+            }
+            vku::FreePnextChain(bind_info.pNext);
+        }
         if (trace_all_semaphores_) {
             LogSubmitInfoSemaphores(submit_info);
         }
